(function($) {
    if (!window.neo) {
        /** @namespace */
        neo = {};
    }
    if (!neo.core) {
        /** @namespace */
        neo.core = {};
    }

    // クラスのショートカット。
    var Pager =
    /**
     * ページャー。
     * .
     * ページの遷移を行う部分を自動で出力する。
     * リンク部分の動作は利用者側で実装する。
     * @class
     */
    neo.core.Pager = function() {
        return this.__constructor.apply(this, arguments);
    };

    // Class Properties:
    $.extend(Pager, /** @lends neo.core.Pager */ {
        /**
         * デフォルトオプション。
         * @type Map
         */
        defaults: {
            current: 1,
            forceLinking: false,
            length: 1,
            max: 10,
            url: '#page-$page$',
            classNames: {
                'prev': 'neo-pager-prev',
                'paging': 'neo-pager-paging',
                'current': 'neo-pager-current',
                'next': 'neo-pager-next',
                'disabled': 'neo-pager-disabled'
            },
            messages: {
                'prev': '<< prev',
                'paging': '$page$',
                'next': 'next >>'
            },
            onclick: null
        }
    });

    // Instance Properties:
    $.extend(Pager.prototype, /** @lends neo.core.Pager.prototype */ {
        /**
         * オプション。
         * @type Map
         */
        options: null,

        /**
         * 出力先の要素。
         */
        _elSelf: null,
        _getElSelf: function() { return this._elSelf; },
        _setElSelf: function(obj) { return this._elSelf = obj; },

    // method:

        /**
         * コンストラクター。
         * @param {HTMLElement} elSelf 出力先のHTML要素。
         * @param {Map} [options] オプション。
         */
        __constructor: function(elSelf, options) {
            // 本体要素を保存
            this._setElSelf(elSelf);

            // オプションを保存
            this._initOptions(options);
        },

        /**
         * オプションを初期化する。
         * .
         * {@link #__constructor}()から呼ばれる。
         * @param {Map} options クラスに与えられたオプション。
         */
        _initOptions: function(options) {
            this.options = this._getComposedOptions(options);
        },

        /**
         * オプションをデフォルト値と合成したものを返す。
         * .
         * {@link #_initOptions}()から呼ばれる。
         * @param {Map} optionsOrg クラスに与えられたオプション。
         * @returns {Map} デフォルト値と合成したオプション。
         * @see neo.core.Pager.defaults
         */
        _getComposedOptions: function(optionsOrg) {
            var options = $.extend({}, Pager.defaults, optionsOrg);
            return options;
        },

        /**
         * 初期処理。
         */
        initialize: function() {
            // 出力
            this.output();
        },

        /**
         * ページャーを出力する。
         * .
         * {@link #initialize}()から呼ばれる。
         */
        output: function() {
            // 出力先要素を取得
            var elSelf = this._getElSelf();

            // 出力要素を取得
            var current = this.options.current;
            var length = this.options.length;
            var elOutput = this._getOutput(current, length);

            // 内容を差し替え
            elSelf.
                empty().
                append(elOutput);

            // フック
            this._hook(elSelf);
        },

        /**
         * 出力要素を返す。
         * .
         * {@link #output}()から呼ばれる。
         * <br>
         * {@link #_getElBefore}(), {@link #_getElPaging}(), {@link #_getElAfter}() を順に呼び、
         * その戻り値の要素をまとめひとつのオブジェクトにしたものが戻り値。
         * @param {Number} current 表示中のページ。
         * @param {Number} length ページ数。
         * @return {HTMLElement} 出力する要素。数は一つとは限らない。
         */
        _getOutput: function(current, length) {
            var max = this.options.max;
            var forced = this.options.forceLinking;

            // ページ番号より前
            var elBefore = this._getElBefore(current, length, forced);

            // ページ番号
            var elPaging = this._getElPaging(current, length, max, forced);

            // ページ番号より後
            var elAfter = this._getElAfter(current, length, forced);

            // 一つのオブジェクトにまとめる
            var elOutput = $('empty').
                add(elBefore).
                add(elPaging).
                add(elAfter);

            return elOutput;
        },

        /**
         * ページ番号より前部分の要素を返す。
         * .
         * {@link #_getOutput}()から呼ばれる。
         * @param {Number} current 表示中のページ。
         * @param {Number} length ページ数。
         * @param {Boolean} forced 強制的にリンクにするフラグ。
         * @return {HTMLElement} ページ番号の前に出力する要素。
         */
        _getElBefore: function(current, length, forced) {
            var el; // 返す要素

            // 表示文言
            var msg = this.options.messages['prev'];

            // リンクでない場合、span で出力
            if (current == 1 && !forced) {
                el = $('<span />');
            }
            // リンクがある場合、a で出力
            else {
                // リンク先
                var url = this._getUrl(Math.min(current-1, length));

                // 要素を生成
                el = $('<a />').attr({'href': url});
            }

            // クラス名とテキストを設定
            el.
                addClass(this.options.classNames['prev']).
                text(msg);

            // 先頭のページを表示している場合、利用不可の状態に
            if (current == 1) {
                el.addClass(this.options.classNames['disabled']);
            }

            return el;
        },

        /**
         * ページ番号より後部分の要素を返す。
         * .
         * {@link #_getOutput}()から呼ばれる。
         * @param {Number} current 表示中のページ。
         * @param {Number} length ページ数。
         * @param {Boolean} forced 強制的にリンクにするフラグ。
         * @return {HTMLElement} ページ番号の後に出力する要素。
         */
        _getElAfter: function(current, length, forced) {
            var el; // 返す要素

            // 表示文言
            var msg = this.options.messages['next'];

            // リンクでない場合、span で出力
            if (current == length && !forced) {
                el = $('<span />');
            }
            // リンクがある場合、a で出力
            else {
                // リンク先
                var url = this._getUrl(Math.min(current+1, length));

                // 要素を生成
                el = $('<a />').attr({'href': url});
            }

            // クラス名とテキストを設定
            el.
                addClass(this.options.classNames['next']).
                text(msg);

            // 末尾のページを表示している場合、利用不可の状態に
            if (current == length) {
                el.addClass(this.options.classNames['disabled']);
            }

            return el;
        },

        /**
         * ページ番号の部分の要素を返す。
         * .
         * {@link #_getOutput}()から呼ばれる。
         * @param {Number} current 表示中のページ。
         * @param {Number} length ページ数。
         * @param {Number} max 最大出力件数。
         * @param {Boolean} forced 強制的にリンクにするフラグ。
         * @return {HTMLElement} ページ番号として出力する要素。
         */
        _getElPaging: function(current, length, max, forced) {
            var el = $('empty');    // 返す要素

            // 開始ページを取得
            var start = this._getPagingStart(current, length, max);

            // 終了ページを取得
            var end = this._getPagingEnd(start, length, max);

            // ページ数分の要素を取得
            for (var i = start; i <= end; i++) {
                var elOne = this._getElPagingOne(i, current, forced);
                el = el.add(elOne);
            }

            return el;
        },

        /**
         * 出力するページ番号の最初を返す。
         * .
         * {@link #_getElPaging}()から呼ばれる。
         * @param {Number} current 表示中のページ。
         * @param {Number} length ページ数。
         * @param {Number} max 最大出力件数。
         * @returns {Number} 開始ページ。
         */
        _getPagingStart: function(current, length, max) {
            var start;

            // 最大出力件数に収まる場合、最初から
            if (length <= max) {
                start = 1;
            }
            // 最大出力件数に収まらない場合、
            // 表示ページが中央に来る位置から
            else {
                start = current - parseInt(max / 2);

                // 最低でも1
                if (start < 1) {
                    start = 1;
                }
                // 最大でも、右端が末尾ページ
                else if (start > length - max + 1) {
                    start = length - max + 1;
                }
            }

            return start;
        },

        /**
         * 出力するページ番号の最初を返す。
         * .
         * {@link #_getElPaging}()から呼ばれる。
         * @param {Number} start 開始ページ。
         * @param {Number} length ページ数。
         * @param {Number} max 最大出力件数。
         * @returns {Number} 開始ページ。
         */
        _getPagingEnd: function(start, length, max) {
            var end;

            end = Math.min(start + max - 1, length);

            return end;
        },

        /**
         * ページ番号の部分のひとつの要素を返す。
         * .
         * {@link #_getElPaging}()から呼ばれる。
         * @param {Number} page ページ番号。
         * @param {Number} current 表示中のページ。
         * @param {Boolean} forced 強制的にリンクにするフラグ。
         * @return {HTMLElement} ページ番号として出力する要素。
         */
        _getElPagingOne: function(page, current, forced) {
            var el; // 返す要素

            // 表示文言
            var template = this.options.messages['paging'];
            var msg = this._getPagingText(template, page);

            // リンクでない場合、span で出力
            if (page == current && !forced) {
                el = $('<span />');
            }
            // リンクがある場合、a で出力
            else {
                // リンク先
                var url = this._getUrl(page);

                // 要素を生成
                el = $('<a />').attr({'href': url});
            }

            // クラス名とテキストを設定
            el.
                addClass(this.options.classNames['paging']).
                text(msg);

            // これが表示中のものである場合、利用不可の状態に
            if (page == current) {
                el.addClass(this.options.classNames['current']);
            }

            return el;
        },

        /**
         * ページャーの要素をフックする。
         * .
         * {@link #output}()から呼ばれる。
         * @param {HTMLElement} elContainer ページャーを括る要素。
         */
        _hook: function(elContainer) {
            // コールバック関数が指定されている場合、フックする
            if (typeof this.options.onclick == 'function') {
                elContainer.find('a').click(this.__closure(this.onclick));
            }
        },

        /**
         * リンク先を返す。
         */
        _getUrl: function(page) {
            var template = this.options.url;
            var url = this._getPagingText(template, page);
            return url;
        },

        /**
         * 任意のテンプレートにページ番号を埋め込んで返す。
         * @example
         * var template = 'list.cgi?page=$page$';
         * var page = 3;
         * var url = this._getPagingText(template, page); // => 'list.cgi?page=3'
         * @param {String} template 元の文字列。この中に含まれる"$page$"をページ番号に置換する。
         * @param {Number} page ページ番号。置換時にゼロ詰めは行われない。
         * @return {String} 置換後の文字列。
         */
        _getPagingText: function(template, page) {
            // 置換
            var msg = template.replace(/\$page\$/g, page);

            return msg;
        },

        /**
         * クリックされたリンク要素から、ページ番号を返す。
         * .
         * {@link #onclick}() から呼ばれる。
         * @param {HTMLElement} elLink リンク要素。
         * @param {String} template URLのテンプレート。
         */
        _getPageOfLink: function(elLink, template) {
            var FIELD_PAGE = '$page$';

            // テンプレートの埋め込みキーワード前後を切りだす準備
            var indexPage = template.indexOf(FIELD_PAGE);
            var prefix = template.slice(indexPage + FIELD_PAGE.length);

            // ページ番号終了位置を取得
            var url = elLink.attr('href');
            var indexPrefix;
            if (prefix == '') {
                indexPrefix = url.length;
            }
            else {
                indexPrefix = url.indexOf(prefix, indexPage);
            }

            // ページ部分を抜き出す
            var pageText = url.slice(indexPage, indexPrefix);
            var page = parseInt(pageText) || 0;

            return page;
        },

        // HTML要素を返す。
        toHtml: function() { return this._getElSelf(); },

        // クロージャーを返す。
        __closure: function(f) {
            var self = this;
            return function() {
                // 引数末尾に実行コンテキストを追加
                var params = Array.prototype.slice.apply(arguments);
                params.push(this);
                return f.apply(self, params);
            };
        },

    // listener:

        /**
         * ページングのリンクをクリックした際の処理。
         * .
         * {@link #options}.onclick にコールバック関数が指定されている場合にのみ割り当てられる。
         * <br>
         * コールバック関数へは、{@link #_getPageOfLink}() で取得したページ番号を渡す。
         * @param {Event} event イベント。
         * @param {RawHTMLElement} elLinkRaw イベント発生要素。
         * @see #_hook
         */
        onclick: function(event, elLinkRaw) {
            // イベントの伝播を遮断
            event.preventDefault();

            // ページ番号を取得
            var elLink = $(elLinkRaw);
            var template = this.options.url;
            var page = this._getPageOfLink(elLink, template);

            // コールバック関数を実行
            this.options.onclick.apply(elLinkRaw, [event, page]);
        }
    });

    jQuery.fn.neoPager = function(options) {
        return this.each(function(i, elRaw) {
            var el = $(elRaw);
            var pager = new Pager(el, options);
            pager.initialize();
        });
    };
})(jQuery);
